import warnings
from pathlib import Path

import numpy as np
import rasterio

from easy_raster.transform.normalizer import Normalizer

warnings.filterwarnings("ignore", module='rasterio')


class RasterIO:
    verbose = False

    @staticmethod
    def read(path: Path, band: int = None):
        with rasterio.open(str(path)) as _:
            data = _.read()
        if band is not None:
            data = data[band]
        return data

    @staticmethod
    def read_normalized(path: Path, band: int = None):
        return Normalizer.normalize(RasterIO.read(path, band))

    @staticmethod
    def write(path: Path | str, buffer: np.ndarray):
        path = Path(path)
        if RasterIO.verbose:
            from loguru import logger
            logger.debug(f'saving buffer {buffer.min():.3f}-{buffer.max():.3f} to {path}')

        path.parent.mkdir(parents=True, exist_ok=True)
        if buffer.ndim < 3:
            buffer = buffer[np.newaxis, :, :]
        depth, h, w = buffer.shape

        if buffer.dtype not in [np.uint8, np.uint16] and path.suffix.lower() in ['.png']:
            raise ValueError(
                'PNG only support uint8 or uint16. Tips: ImageIO.write_normalize() will handle float data type.'
            )

        with rasterio.open(str(path), 'w', width=w, height=h, count=depth, dtype=buffer.dtype) as _:
            _.write(buffer, indexes=[1 + i for i in range(depth)])

    @staticmethod
    def write_normalize(path: Path | str, data: np.ndarray):
        path = Path(path)
        data = Normalizer.normalize(data)

        if path.suffix == '.png':
            data = (data * 255).astype(np.uint8)

        RasterIO.write(path, data)



from dataclasses import dataclass

import numpy as np

from easy_raster.model.position import Position


@dataclass
class Resolution:
    width: int = 128
    height: int = 128

    @staticmethod
    def square(size: int):
        return Resolution.from_raw(size, size)

    @staticmethod
    def from_raw(height: int, width: int):
        return Resolution(width=width, height=height)

    def rescale_size(self, size: float) -> int:
        assert 0 <= size <= 1
        return int(size * self.width)

    def rescale_position(self, position: Position):
        return int(position.y * self.width), int(position.x * self.height)

    def area(self):
        return self.width * self.height

    def new_buffer(self, bands: int = None, dtype=None):
        if bands is None:
            return np.zeros((self.height, self.width), dtype=dtype)

        return np.zeros((self.height, self.width, bands))

    def raw(self):
        return self.height, self.width

    def __repr__(self):
        return f'{self.width}x{self.height}'

    def __str__(self):
        return repr(self)



'''
Module for fast morphological operations for binary images
'''

__all__ = ['binary_dilation', 'binary_erosion', 'binary_closing', 'binary_opening']

import numpy as np

def binary_dilation(image, selem=None, out=None):
    # https://github.com/shoheiogawa/binmorphopy


    '''Apply a binary-dilation operation to an image using a structuring element

    This function removes pixels to objects' perimeter in the image using
    a structuring element.

    Args:
        image (array-like):
            Image data as an array. If the input is not numpy.bool array,
            the data is converted to this type.
        selem (array-like):
            Structuring element as an boolean image of the same dimension of `image`
        out (numpy.bool array):
            Array to store the result of this operation. The length of the array
            must be the same as the input image.

    Returns:
        numpy.bool array: Dilated image (when `out` is `None`)
    '''
    dim = image.ndim
    if not isinstance(image, np.ndarray):
        image = np.asarray(image)
    if not image.dtype == bool:
        image = image.astype(bool)
    if selem is None:
        if dim == 1:
            selem = np.ones(shape=[3], dtype=bool)
        elif dim == 2:
            selem = np.zeros(shape=[3, 3], dtype=bool)
            selem[1, :] = True
            selem[:, 1] = True
        elif dim == 3:
            selem = np.zeros(shape=[3, 3, 3], dtype=bool)
            selem[:, 1, 1] = True
            selem[1, :, 1] = True
            selem[1, 1, :] = True
    else:
        if not isinstance(selem, np.ndarray):
            selem = np.asarray(selem, dtype=bool)
        if not selem.dtype == bool:
            selem = selem.astype(bool)
        if any([num_pixels % 2 == 0 for num_pixels in selem.shape]):
            raise ValueError('Only structure element of odd dimension '
                             'in each direction is supported.')
    perimeter_image = _get_perimeter_image(image)
    perimeter_coords = np.where(perimeter_image)
    if out is None:
        return_out = True
        out = image.copy()
    else:
        return_out = False
        out[:] = image[:]

    if dim == 1:
        sx = selem.shape[0]
        rx = sx // 2
        lx = image.shape[0]
        for ix in perimeter_coords[0]:
            (jx_b, jx_e), (kx_b, kx_e) = _generate_array_indices(ix, rx, sx, lx)
            out[jx_b:jx_e] |= selem[kx_b:kx_e]

    if dim == 2:
        rx, ry = [n // 2 for n in selem.shape]
        lx = image.shape
        sx, sy = selem.shape
        lx, ly = image.shape
        for ix, iy in zip(perimeter_coords[0], perimeter_coords[1]):
            (jx_b, jx_e), (kx_b, kx_e) = _generate_array_indices(ix, rx, sx, lx)
            (jy_b, jy_e), (ky_b, ky_e) = _generate_array_indices(iy, ry, sy, ly)
            out[jx_b:jx_e, jy_b:jy_e] |= selem[kx_b:kx_e, ky_b:ky_e]

    if dim == 3:
        rx, ry, rz = [n // 2 for n in selem.shape]
        sx, sy, sz = selem.shape
        lx, ly, lz = image.shape
        for ix, iy, iz in zip(perimeter_coords[0], perimeter_coords[1], perimeter_coords[2]):
            (jx_b, jx_e), (kx_b, kx_e) = _generate_array_indices(ix, rx, sx, lx)
            (jy_b, jy_e), (ky_b, ky_e) = _generate_array_indices(iy, ry, sy, ly)
            (jz_b, jz_e), (kz_b, kz_e) = _generate_array_indices(iz, rz, sz, lz)
            out[jx_b:jx_e, jy_b:jy_e, jz_b:jz_e] |= selem[kx_b:kx_e, ky_b:ky_e, kz_b:kz_e]

    if return_out:
        return out

def binary_erosion(image, selem=None, out=None):
    '''Apply a binary-erosion operation to an image using a structuring element

    This function removes pixels around objects' perimeter in an image and returns
    the result as an image.
    See the `binary_dilation` function doc-string for the arguments and retuned value.
    '''
    if not isinstance(image, np.ndarray):
        image = np.asarray(image)
    if not image.dtype == bool:
        image = image.astype(bool)

    out_image = binary_dilation(~image, selem, out)

    if out is None:
        return ~out_image
    else:
        out[:] = ~out[:]

def binary_closing(image, selem=None, out=None):
    '''Apply a binary-closing operation to an image using a structuring element

    This function dilates an image and then erodes the dilation result.
    See the `binary_dilation` function doc-string for the arguments and retuned value.
    '''
    out_image = binary_erosion(binary_dilation(image, selem), selem, out)
    if out is None:
        return out_image

def binary_opening(image, selem=None, out=None):
    '''Apply a binary-opening operation to an image using a structuring element

    This function erodes an image and then dilates the eroded result.
    See the `binary_dilation` function doc-string for arguments and retuned value.
    '''
    out_image = binary_dilation(binary_erosion(image, selem), selem, out)
    if out is None:
        return out_image

def _get_perimeter_image(image):
    '''Return the image of the perimeter structure of the input image

    Args:
        image (Numpy array): Image data as an array

    Returns:
        Numpy array: Perimeter image
    '''
    dim = image.ndim
    if dim > 3:
        raise RuntimeError('Binary image in 4D or above is not supported.')
    count = np.zeros_like(image, dtype=np.uint8)
    inner = np.zeros_like(image, dtype=bool)

    count[1:] += image[:-1]
    count[:-1] += image[1:]

    if dim == 1:
        inner |= image == 2
        for i in [0, -1]:
            inner[i] |= count[i] == 1
        return image & (~inner)

    count[:, 1:] += image[:, :-1]
    count[:, :-1] += image[:, 1:]
    if dim == 2:
        inner |= count == 4
        for i in [0, -1]:
            inner[i] |= count[i] == 3
            inner[:, i] |= count[:, i] == 3
        for i in [0, -1]:
            for j in [0, -1]:
                inner[i, j] |= count[i, j] == 2
        return image & (~inner)

    count[:, :, 1:] += image[:, :, :-1]
    count[:, :, :-1] += image[:, :, 1:]

    if dim == 3:
        inner |= count == 6
        for i in [0, -1]:
            inner[i] |= count[i] == 5
            inner[:, i] |= count[:, i] == 5
            inner[:, :, i] |= count[:, :, i] == 5
        for i in [0, -1]:
            for j in [0, -1]:
                inner[i, j] |= count[i, j] == 4
                inner[:, i, j] |= count[:, i, j] == 4
                inner[:, i, j] |= count[:, i, j] == 4
                inner[i, :, j] |= count[i, :, j] == 4
                inner[i, :, j] |= count[i, :, j] == 4
        for i in [0, -1]:
            for j in [0, -1]:
                for k in [0, -1]:
                    inner[i, j, k] |= count[i, j, k] == 3
        return image & (~inner)
    raise RuntimeError('This line should not be reached.')

def _generate_array_indices(selem_center, selem_radius, selem_length, result_length):
    '''Return the correct indices for slicing considering near-edge regions

    Args:
        selem_center (int): The index of the structuring element's center
        selem_radius (int): The radius of the structuring element
        selem_length (int): The length of the structuring element
        result_length (int): The length of the operating image

    Returns:
        (int, int): The range begin and end indices for the operating image
        (int, int): The range begin and end indices for the structuring element image
    '''
    # First index for the result array
    result_begin = selem_center - selem_radius
    # Last index for the result array
    result_end = selem_center + selem_radius + 1
    # First index for the structuring element array
    selem_begin = -result_begin if result_begin < 0 else 0
    result_begin = max(0, result_begin)
     # Last index for the structuring element array
    selem_end = selem_length -(result_end - result_length) \
                    if result_end > result_length else selem_length
    return (result_begin, result_end), (selem_begin, selem_end)



import numpy as np

from easy_raster.transform.buffer_util import BufferUtil
from easy_raster.transform.normalizer import Normalizer


class RFactory:

    @staticmethod
    def heightmap(size: int, n: int = 5):
        height = np.zeros((size, size))
        weight = 0
        for i in range(n):
            amplitude = 2 ** i
            s = size // amplitude
            x = np.random.random((s, s))
            x = BufferUtil.smooth(x, sigma=3)
            x = BufferUtil.resize(x, width=size, height=size)
            x = 2 * np.abs(x - .5)
            height += x * amplitude
            weight += amplitude

        return Normalizer.normalize(height)

    @staticmethod
    def random_gradient(shape: tuple[int, ...]):
        grad = np.exp(2j * np.pi * np.random.rand(*shape))
        grad = np.dstack([grad.real, grad.imag])
        return np.moveaxis(grad, 2, 0)



from easy_kit.timing import time_func


@time_func
def clamp(x: float, low: float, high: float):
    if x is None:
        return x
    if low is None:
        low = x
    if high is None:
        high = x
    return max(low, min(x, high))


@time_func
def clamp_abs(x: float, low: float, high: float):
    if x is None:
        return x
    if low is None:
        low = abs(x)
    if high is None:
        high = abs(x)
    if abs(x) < low:
        return 0
    if x > high:
        return high
    if x < - high:
        return -high
    return x



from pathlib import Path


def make_video(input_glob: Path | str, output: Path | str, fps: int = 1):
    import ffmpeg

    (
        ffmpeg
        .input(str(input_glob), pattern_type='glob', framerate=fps)
        .output(str(output), loglevel="quiet")
        .overwrite_output()
        .run()
    )



import numpy as np

MAX_PIXEL_VALUE = 255


class BufferUtil:
    @staticmethod
    def mirror_tile(buffer: np.ndarray):
        top = np.hstack([buffer, np.fliplr(buffer)])
        return np.vstack([top, np.flipud(top)])

    @staticmethod
    def normalize(buffer: np.ndarray, png_format: bool = False):
        if buffer.min() != buffer.max():
            buffer -= buffer.min()
        if buffer.max() != 0:
            buffer = buffer / buffer.max()

        if png_format:
            buffer = (buffer * MAX_PIXEL_VALUE).astype('uint8')
        else:
            buffer = buffer.astype('float32')

        return buffer

    @staticmethod
    def resize(buffer: np.ndarray, width: int, height: int, seamless: bool = False):
        import skimage

        res = skimage.transform.resize(buffer, (height, width), anti_aliasing=True)

        if seamless:
            res[-1, :] = res[0, :]
            res[:, -1] = res[:, 0]

        return res

    @staticmethod
    def smooth(buffer: np.ndarray, sigma: float = 1):
        from scipy.ndimage import gaussian_filter

        return gaussian_filter(buffer, sigma=sigma, order=0)



import numpy as np


class Normalizer:
    @staticmethod
    def identity(x: np.ndarray):
        return x

    @staticmethod
    def log_scale(x: np.ndarray):
        y = x - x.min()
        try:
            return np.log10(y + 1)
        except Exception as e:
            print(y)
            raise e

    @staticmethod
    def power(y: float):
        def func(x: np.ndarray):
            return np.power(x, y)

        return func

    @staticmethod
    def normalize(x: np.ndarray):
        a, b = x.min(), x.max()

        if np.any(b == 0):
            return x

        return (x - a) / (b - a)



import random
from dataclasses import dataclass


@dataclass
class Position:
    x: float = 0
    y: float = 0

    @staticmethod
    def random():
        return Position(x=random.random(), y=random.random())

    @staticmethod
    def from_raw(y: float, x: float):
        return Position(x=x, y=y)

    def raw(self):
        return self.y, self.x
